<%
' 2000-2005 L.C. Enterprises
'http://LCen.com
%>
<%
'Bayes Filter Class
'Author: Luis Cantero
'Updated: 06/AUG/2005

'INITIALIZE:	Set objBayesFilter = New BayesFilter, .GoodThreshold, .BadThreshold, .AutoTrain, .SignificantTokens, .UnkownBadProb
'INITIALIZE:	.MinTokenOccurrence, .TokenCharset, .ConnectionObject, .MaxTokenLength, .CompareMode, .DatabaseLoad
'PROPERTIES:	.TokenCount, .GoodCount, .BadCount
'METHODS:		.TrainGood(strText), .TrainBad(strText), .DatabaseLoad(), .DatabaseSave()
'FUNCTIONS:		.Evaluate(strText)
Class BayesFilter '-------------------------- CLASS BEGIN --------------------------

'Initialization properties
Private AUTO_TRAIN_GOOD_THRESHOLD 'As Double	'0.1
Private AUTO_TRAIN_BAD_THRESHOLD 'As Double		'0.9
Private AUTO_TRAIN 'As Boolean					'False
Private SIGNIFICANT_TOKENS 'As Integer			'15
Private UNKOWN_BAD_PROB 'As Double				'0.5
Private MIN_TOKEN_OCCURRENCE 'As Integer		'5
Private TOKEN_CHARSET 'As String
Private MAX_TOKEN_LENGTH 'As Integer			'255
Private COMPARE_MODE 'As Integer				'vbTextCompare (Case Insensitive)
Private objDBConn 'As Object

'Dictionaries (hash tables)
Private dicTokenGoodCount 'As Object
Private dicTokenBadCount 'As Object
Private dicTokenProb 'As Object

'Counters
Private intEventGoodCount 'As Integer
Private intEventBadCount 'As Integer

'Indicates the upper bound for good tokens, for auto-learning purposes
Public Property Let AutoTrainGoodThreshold(dblInput)

	AUTO_TRAIN_GOOD_THRESHOLD = dblInput

End Property

'Indicates the lower bound for bad tokens, for auto-learning purposes
Public Property Let AutoTrainBadThreshold(dblInput)

	AUTO_TRAIN_BAD_THRESHOLD = dblInput

End Property

'Auto train when classifying based on good and bad thresholds
Public Property Let AutoTrain(blnInput)

	AUTO_TRAIN = blnInput

End Property

'Number of tokens to consider when evaluating a string
Public Property Let SignificantTokens(intInput)

	SIGNIFICANT_TOKENS = intInput

End Property

'Prob to assign to an unkown outcome
Public Property Let UnkownBadProb(dblInput)

	UNKOWN_BAD_PROB = dblInput

End Property

'Minimum number of occurrences for a token to be considered
Public Property Let MinTokenOccurrence(intInput)

	MIN_TOKEN_OCCURRENCE = intInput

End Property

'Max string length for the tokens (for database saving)
Public Property Let MaxTokenLength(intInput)

	MAX_TOKEN_LENGTH = intInput

End Property

'Compare mode for the dictionaries
Public Property Let CompareMode(intInput)

	COMPARE_MODE = intInput

End Property

'Allowed charset for tokens, all other (visible) chars are separators
Public Property Let TokenCharset(strInput)

	TOKEN_CHARSET = strInput

End Property

'Database connection object
Public Property Let ConnectionObject(objInput)

	Set objDBConn = objInput

End Property

'Returns the token count
Public Property Get TokenCount()

	TokenCount = dicTokenProb.Count

End Property

'Returns the good event count
Public Property Get GoodCount()

	GoodCount = intEventGoodCount

End Property

'Returns the bad event count
Public Property Get BadCount()

	BadCount = intEventBadCount

End Property

Private Sub Class_Initialize()

	'Default values
	AUTO_TRAIN_GOOD_THRESHOLD = 0.1
	AUTO_TRAIN_BAD_THRESHOLD = 0.9
	AUTO_TRAIN = False
	SIGNIFICANT_TOKENS = 15
	UNKOWN_BAD_PROB = 0.5
	MIN_TOKEN_OCCURRENCE = 5
	MAX_TOKEN_LENGTH = 255
	COMPARE_MODE = vbTextCompare
	TOKEN_CHARSET = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-!$"
	intEventGoodCount = 0
	intEventBadCount = 0

	'vbBinaryCompare = 0 (Case sensitive)
	Set dicTokenGoodCount = CreateObject("Scripting.Dictionary")
	Set dicTokenBadCount = CreateObject("Scripting.Dictionary")
	Set dicTokenProb = CreateObject("Scripting.Dictionary")
	dicTokenProb.CompareMode = COMPARE_MODE

End Sub

Private Sub Class_Terminate()

	'Clean up
	Set objDBConn = Nothing
	Set dicTokenGoodCount = Nothing
	Set dicTokenBadCount = Nothing
	Set dicTokenProb = Nothing

End Sub

'Train text as being good
Public Sub TrainGood(strText)

	Dim dicTemp 'As Object
	Set dicTemp = GetTextDict(strText)

	Call TrainDictGood(dicTemp)

	Set dicTemp = Nothing

End Sub

'Train text as being bad
Public Sub TrainBad(strText)

	Dim dicTemp 'As Object
	Set dicTemp = GetTextDict(strText)

	Call TrainDictBad(dicTemp)

	Set dicTemp = Nothing

End Sub

'Implementation for TrainDict (good)
Private Sub TrainDictGood(dicTable)

	Call TrainDict(dicTable, 0)

End Sub

'Implementation for TrainDict (bad)
Private Sub TrainDictBad(dicTable)

	Call TrainDict(dicTable, 1)

End Sub

'0 = Train as good / 1 = Train as bad
Private Sub TrainDict(dicTable, intTrainType)

	Dim arrKeys
	Dim dicTrain 'As Object
	Dim dicOther 'As Object
	Dim i 'As Long

	Select Case intTrainType
		Case 0 'Good
			Set dicTrain = dicTokenGoodCount
			Set dicOther = dicTokenBadCount
			intEventGoodCount = intEventGoodCount + 1

		Case 1 'Bad
			Set dicTrain = dicTokenBadCount
			Set dicOther = dicTokenGoodCount
			intEventBadCount = intEventBadCount + 1

	End Select

	arrKeys = dicTable.Keys()

	'Get all tokens and their prob and put them in a dictionary
	'Increase count by one regardless of number of ocurrences of the token in the text
	For i = 0 To UBound(arrKeys)
		If dicTokenProb.Exists(arrKeys(i)) Then
			dicTrain(arrKeys(i)) = dicTrain(arrKeys(i)) + 1
		Else 'Add new token
			Call dicTrain.Add(arrKeys(i), 1)
			Call dicOther.Add(arrKeys(i), 0)
			Call dicTokenProb.Add(arrKeys(i), 0)
		End If
	Next

	'Update Bayes probabilities for all tokens, necessary because changes in event count affect all probabilities
	Call UpdateBayProb

End Sub

'Classifies text by calculating its probability of being bad
Public Function Evaluate(strText)

	Dim arrKeys
	Dim dicText 'As Object
	Dim dicSignificant 'As Object
	Dim dblTemp 'As Double
	Dim i 'As Long

	Set dicText = GetTextDict(strText)
	Set dicSignificant = GetMostSignificant(dicText)

	arrKeys = dicSignificant.Keys()

	'Add all probabilities
	For i = 0 To UBound(arrKeys)
		dblTemp = dblTemp + dicSignificant(arrKeys(i))
	Next

	'Return
	If (UBound(arrKeys) + 1) > 0 Then
		Evaluate = dblTemp / (UBound(arrKeys) + 1)
	Else
		Evaluate = UNKOWN_BAD_PROB
	End If

	'Auto train only when enabled and if the minimum amount of significant tokens have been found
	If AUTO_TRAIN And dicSignificant.Count = SIGNIFICANT_TOKENS Then
		If Evaluate >= AUTO_TRAIN_BAD_THRESHOLD Then 'Train as bad
			Call TrainDict(dicText, 1)
		Else
			If Evaluate <= AUTO_TRAIN_GOOD_THRESHOLD Then 'Train as good
				Call TrainDict(dicText, 0)
			End If
		End If
	End If

	'Clean up
	Set dicText = Nothing
	Set dicSignificant = Nothing

End Function

'Returns most significant tokens, measured by their probability of being bad
Private Function GetMostSignificant(dicText)

	Dim arrKeys
	Dim arrTemp()
	Dim dicTemp 'As Object
	Dim i 'As Long
	Dim x 'As Long

	Set dicTemp = CreateObject("Scripting.Dictionary")

	arrKeys = dicText.Keys()

	'Consider only tokens that occured a minimum amount of times
	For i = 0 To UBound(arrKeys)
		If dicTokenProb.Exists(arrKeys(i)) Then
			'Token does not meet the minimum occurrece requirement, remove it
			If dicTokenGoodCount(arrKeys(i)) + dicTokenBadCount(arrKeys(i)) < MIN_TOKEN_OCCURRENCE Then
				Call dicText.Remove(arrKeys(i))
			End If
		Else 'Unkown token, remove it
			Call dicText.Remove(arrKeys(i))
		End If
	Next

	'Update key array
	arrKeys = dicText.Keys()

	'Make a list of the most significant (bad) tokens
	If UBound(arrKeys) + 1 > SIGNIFICANT_TOKENS Then
		ReDim arrTemp(SIGNIFICANT_TOKENS, 1)

		'Add first SIGNIFICANT_TOKENS tokens to temp array
		For i = 0 To SIGNIFICANT_TOKENS - 1
			arrTemp(i, 0) = arrKeys(i) 'Token
			arrTemp(i, 1) = dicText(arrKeys(i)) 'Prob
		Next

		'Search for more significant tokens
		For i = SIGNIFICANT_TOKENS To UBound(arrKeys)
			'Compare them one by one with the tokens already in the array
			For x = 0 To SIGNIFICANT_TOKENS - 1
				'Token has higher probability of being bad, use it instead
				If dicText(arrKeys(i)) > arrTemp(x, 1) Then
					arrTemp(x, 0) = arrKeys(i)
					arrTemp(x, 1) = dicText(arrKeys(i))
					Exit For
				End If
			Next
		Next

		'Make a dictionary
		For i = 0 To SIGNIFICANT_TOKENS - 1
			Call dicTemp.Add(arrTemp(i, 0), arrTemp(i, 1))
		Next
	Else
		Set dicTemp = dicText
	End If

	'Return
	Set GetMostSignificant = dicTemp

End Function

'Returns a dictionary containing all found tokens and their prob
Private Function GetTextDict(ByVal strText)

	Dim dicTemp 'As Object
	Dim strTemp 'As String
	Dim intBeg 'As Long
	Dim dblProb 'As Double
	Dim i 'As Long

	Set dicTemp = CreateObject("Scripting.Dictionary")

	dicTemp.CompareMode = COMPARE_MODE

	'Clean up text
	strText = Replace(strText, vbCrLf, " ")

	'Remove invisible chars
	For i = 0 To 31
		strText = Replace(strText, Chr(i), "")
	Next

	intBeg = 1
	For i = 1 To Len(strText)
		If InStr(1, TOKEN_CHARSET, Mid(strText, i, 1)) = 0 Then 'Separator found
			strTemp = Mid(strText, intBeg, i - intBeg)

			'Add token and its probability
			If strTemp <> "" Then
				If Not dicTemp.Exists(strTemp) Then 'Token not yet in list
					If dicTokenProb.Exists(strTemp) Then 'Known token
						dblProb = dicTokenProb(strTemp)
					Else 'Unknown token
						dblProb = UNKOWN_BAD_PROB
					End If

					Call dicTemp.Add(strTemp, dblProb) 'Add with prob
				End If
			End If

			intBeg = i + 1
		End If
	Next

	'Return
	Set GetTextDict = dicTemp

End Function

'Recalculates the Bayes probability of something bad ocurring, updates the whole table
Private Sub UpdateBayProb()

	Dim arrKeys
	Dim dblPBAN 'As Double
	Dim dblPBA 'As Double
	Dim dblPAN 'As Double
	Dim dblPA 'As Double
	Dim dblCorrectionFactor 'As Double
	Dim i 'As Long

	arrKeys = dicTokenProb.Keys()

	dblPAN = GetProbGood(intEventGoodCount, intEventBadCount)
	dblPA = 1 - dblPAN 'Prob bad

	'Correct prob good
	'To avoid masses of bad events decreasing the good event probability too much
	If intEventBadCount > intEventGoodCount Then
		If dblPAN > 0 Then dblCorrectionFactor = (dblPA / dblPAN) / 2
		If dblCorrectionFactor > 0 Then dblPAN = dblPAN * dblCorrectionFactor
	End If

	For i = 0 To UBound(arrKeys)
		'Prob that a token is good
		dblPBAN = GetProbGood(dicTokenGoodCount(arrKeys(i)), dicTokenBadCount(arrKeys(i)))
		'Prob that a token is bad
		dblPBA = 1 - dblPBAN

		dicTokenProb(arrKeys(i)) = CalcBayes(dblPBA, dblPA, dblPBAN, dblPAN)
	Next

End Sub

'Prob of something good, taking into account the number of good and bad experiences
Private Function GetProbGood(intEventGoodCount, intEventBadCount)

	If intEventGoodCount > 0 Or intEventBadCount > 0 Then
		GetProbGood = intEventGoodCount / (intEventGoodCount + intEventBadCount)
	Else 'No data
		GetProbGood = UNKOWN_BAD_PROB
	End If

End Function

'dblPBA: Prob that a token is bad
'dblPA: Prob of bad ocurring
'dblPBAN: Prob that a token is good
'dblPAN: Prob of good occuring
Private Function CalcBayes(dblPBA, dblPA, dblPBAN, dblPAN)

	'Return
	CalcBayes = dblPBA * dblPA / (dblPBA * dblPA + dblPBAN * dblPAN)

End Function

'Load all tokens from a database
Public Sub DatabaseLoad()

	Dim objRecordSet 'As Object
	Dim strSQL 'As String

	'Load tokens, counts and probabilities from a database
	strSQL = "SELECT * FROM tbl_Tokens"
	Set objRecordSet = objDBConn.Execute(strSQL)

	While Not objRecordSet.EOF
		Call dicTokenGoodCount.Add(CStr(objRecordSet("tok_Name")), CLng(objRecordSet("tok_CountGood")))
		Call dicTokenBadCount.Add(CStr(objRecordSet("tok_Name")), CLng(objRecordSet("tok_CountBad")))
		Call dicTokenProb.Add(CStr(objRecordSet("tok_Name")), CDbl(objRecordSet("tok_BayProb")))

		objRecordSet.MoveNext
	Wend

	Set objRecordSet = Nothing

	'Load counters
	strSQL = "SELECT * FROM tbl_Counter"
	Set objRecordSet = objDBConn.Execute(strSQL)

	If Not objRecordSet.EOF Then
		intEventGoodCount = objRecordSet("cnt_Good")
		intEventBadCount = objRecordSet("cnt_Bad")
	End If

	Set objRecordSet = Nothing

End Sub

'Save all tokens to a database
Public Sub DatabaseSave()

	Dim arrKeys
	Dim objRecordSet 'As Object
	Dim strSQL 'As String
	Dim i 'As Long

	'Clear tables
	strSQL = "DELETE FROM tbl_Tokens"
	Call objDBConn.Execute(strSQL)
	strSQL = "DELETE FROM tbl_Counter"
	Call objDBConn.Execute(strSQL)

	Set objRecordSet = CreateObject("ADODB.Recordset")

	Call objRecordSet.Open("tbl_Tokens", objDBConn, 0, 3) 'adOpenForwardOnly, adLockOptimistic

	arrKeys = dicTokenProb.Keys()

	'Save tokens, counts and probabilities to a database
	For i = 0 To UBound(arrKeys)
		objRecordSet.AddNew
		If Len(arrKeys(i)) > MAX_TOKEN_LENGTH Then arrKeys(i) = Left(arrKeys(i), MAX_TOKEN_LENGTH)
		objRecordSet("tok_Name") = arrKeys(i)
		objRecordSet("tok_CountGood") = dicTokenGoodCount(arrKeys(i))
		objRecordSet("tok_CountBad") = dicTokenBadCount(arrKeys(i))
		objRecordSet("tok_BayProb") = dicTokenProb(arrKeys(i))

		objRecordSet.Update
	Next

	Call objRecordSet.Close

	Set objRecordSet = Nothing

	'Save counters
	strSQL = "INSERT INTO tbl_Counter(cnt_Good, cnt_Bad) VALUES(" & intEventGoodCount & ", " & intEventBadCount & ")"
	Call objDBConn.Execute(strSQL)

End Sub

End Class '-------------------------- CLASS END --------------------------
%>